from enum import Enum

from django.conf import settings as django_settings
from django.urls import reverse

import msal
import requests
from onelogin.saml2.auth import OneLogin_Saml2_Auth
from onelogin.saml2.idp_metadata_parser import OneLogin_Saml2_IdPMetadataParser

from auth_token.config import settings


graph_url = 'https://graph.microsoft.com/v1.0'


class Protocol(Enum):
    OAUTH = 'oauth'
    SAML = 'saml'


def get_user_data(token):
    response = requests.get(
        f'{graph_url}/me',
        headers={'Authorization': f'Bearer {token}'},
        params={'$select': 'displayName,mail,userPrincipalName'}
    )
    if response.status_code == 200:
        return response.json()
    else:
        return None


def get_msal_app():
    """
    Initialize the MSAL confidential client
    """
    return msal.PublicClientApplication(
        settings.MS_SSO_APP_ID,
        authority=f'https://login.microsoftonline.com/{settings.MS_SSO_TENANT_ID}'
    )


def get_sign_in_flow():
    """
    Method to generate a sign-in flow
    """
    return get_msal_app().initiate_auth_code_flow(['user.read'])


def acquire_token_by_auth_code_flow(sign_flow, data):
    """
    Method to get auth code from sign flow and request data
    """
    return get_msal_app().acquire_token_by_auth_code_flow(sign_flow, data)


def parse_request_for_saml(request):
    """
    Inspired from the example project:
    https://github.com/SAML-Toolkits/python3-saml/blob/v1.16.0/demo-django/demo/views.py#L17
    """
    return {
        'https': 'on' if request.is_secure() else 'off',
        'http_host': request.get_host(),
        'script_name': request.get_full_path(),
        'get_data': request.GET.copy(),
        'post_data': request.POST.copy(),
    }


def init_saml_auth(request):
    service_provider_settings = {
        'strict': True,
        'debug': django_settings.DEBUG,
        'security': {
            'requestedAuthnContext': False,  # do not enforce any particular authentication method
            **({'allowSingleLabelDomains': True} if getattr(django_settings, "AUTH_TOKEN_TEST", False) else {}),
        },
        'sp': {
            'entityId': settings.MS_SSO_SAML_ENTITY_ID,
            'assertionConsumerService': {
                'url': request.scheme + '://' + request.get_host() + reverse('ms-sso-saml-redirect'),
                'binding': 'urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST',
            },
        },
    }
    identity_provider_settings = OneLogin_Saml2_IdPMetadataParser.parse_remote(settings.MS_SSO_SAML_METADATA_URL)
    merged_settings = OneLogin_Saml2_IdPMetadataParser.merge_settings(
        service_provider_settings,
        identity_provider_settings,
    )
    return OneLogin_Saml2_Auth(parse_request_for_saml(request), merged_settings)


def get_ms_sso_login_url():
    protocol = Protocol(settings.MS_SSO_PROTOCOL)
    if protocol == Protocol.OAUTH:
        return reverse('ms-sso-login')
    elif protocol == Protocol.SAML:
        return reverse('ms-sso-saml-login')



from datetime import timedelta

from django.conf import settings as django_settings
from django.contrib.auth.hashers import check_password, make_password
from django.contrib.auth.models import AnonymousUser
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models, IntegrityError, transaction
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from enumfields import IntegerEnumField

from chamber.exceptions import PersistenceException
from chamber.models import SmartModel, SmartQuerySet, SmartManager

from generic_m2m_field.models import GenericManyToManyField

from auth_token.utils import compute_expires_at, generate_key, hash_key
from auth_token.config import settings

from .enums import AuthorizationRequestState, AuthorizationRequestResult
from .exceptions import KeyGeneratorError


KEY_SALT = 'django-auth-token'


def compute_authorization_token_expires_at(expiration=None):
    return compute_expires_at(expiration or settings.DEFAULT_TOKEN_AGE)


def generate_token_key():
    return generate_key(length=settings.LENGTH)


class BaseHashKeyManager(SmartManager):

    def _hash_key(self, key, **kwargs):
        return hash_key(key)

    def create(self, key_generator, **kwargs):
        for _i in range(settings.MAX_RANDOM_KEY_ITERATIONS):
            try:
                key = key_generator()
                hashed_key = self._hash_key(key, **kwargs)
                obj = super().create(key=hashed_key, **kwargs)
                obj.secret_key = key
                return obj
            except IntegrityError:
                pass
            except PersistenceException as ex:
                if not ex.error_dict and 'key' not in ex.error_dict:
                    raise ex
        raise KeyGeneratorError('Could not produce unique key')


class AuthorizationTokenManager(BaseHashKeyManager):

    def create(self, **kwargs):
        return super().create(generate_token_key, **kwargs)


class AuthorizationToken(SmartModel):
    """
    The default authorization token model.
    """

    key = models.CharField(
        verbose_name=_('key'),
        max_length=128,
        unique=True,
        db_index=True,
        null=False,
        blank=False
    )
    user = models.ForeignKey(
        verbose_name=_('user'),
        to=django_settings.AUTH_USER_MODEL,
        related_name='authorization_tokens',
        null=False,
        blank=False,
        on_delete=models.CASCADE
    )
    is_active = models.BooleanField(
        verbose_name=_('is active'),
        default=True
    )
    user_agent = models.CharField(
        verbose_name=_('user agent'),
        max_length=256,
        null=True,
        blank=True
    )
    expires_at = models.DateTimeField(
        verbose_name=_('expires at'),
        null=False,
        blank=False,
        default=compute_authorization_token_expires_at
    )
    ip = models.GenericIPAddressField(
        verbose_name=_('IP'),
        null=False,
        blank=False
    )
    auth_slug = models.SlugField(
        verbose_name=_('slug'),
        null=True,
        blank=True
    )
    backend = models.CharField(
        verbose_name=_('backend'),
        max_length=250,
        null=False,
        blank=False
    )
    allowed_cookie = models.BooleanField(
        verbose_name=_('is allowed cookie'),
        default=True
    )
    allowed_header = models.BooleanField(
        verbose_name=_('is allowed header'),
        default=True
    )
    is_authenticated = models.BooleanField(
        verbose_name=_('is authenticated'),
        null=False,
        blank=False,
        default=False
    )
    preserve_cookie = models.BooleanField(
        verbose_name=_('preserve cookie'),
        null=False,
        blank=False,
        default=False
    )
    mobile_device = models.ForeignKey(
        verbose_name=_('mobile device'),
        to='MobileDevice',
        related_name='authorization_tokens',
        null=True,
        blank=True,
        on_delete=models.CASCADE
    )

    related_objects = GenericManyToManyField()

    is_from_header = False
    is_from_cookie = False
    secret_key = None
    is_anonymous = False

    objects = AuthorizationTokenManager()

    class Meta:
        verbose_name = _('authorization token')
        verbose_name_plural = _('authorization tokens')

    class SmartMeta:
        is_cleaned_pre_save = False

    @cached_property
    def active_takeover(self):
        return self.user_takeovers.filter(is_active=True).last('created_at', 'pk')

    @property
    def is_expired(self):
        return self.expires_at < timezone.now()

    @property
    def time_to_expiration(self):
        return max(timedelta(seconds=0), self.expires_at - timezone.now())


class AnonymousAuthorizationToken:

    key = None
    user = AnonymousUser()
    created_at = None
    is_active = False
    user_agent = None
    expiration = None
    is_from_header = False
    is_from_cookie = False
    active_takeover = None
    backend = None
    allowed_cookie = False
    allowed_header = False
    secret_key = None
    is_authenticated = False
    is_anonymous = True

    def save(self):
        raise NotImplementedError

    def delete(self):
        raise NotImplementedError


class UserAuthorizationTokenTakeover(SmartModel):
    """
    The model allows to change user without token change
    """

    token = models.ForeignKey(
        verbose_name=_('authorization token'),
        to=AuthorizationToken,
        related_name='user_takeovers',
        on_delete=models.CASCADE
    )
    user = models.ForeignKey(
        verbose_name=_('user'),
        to=django_settings.AUTH_USER_MODEL,
        related_name='user_token_takeovers',
        null=False,
        blank=False,
        on_delete=models.CASCADE
    )
    is_active = models.BooleanField()

    class Meta:
        verbose_name = _('authorization takeover')
        verbose_name_plural = _('authorization takeovers')


class MobileDeviceAlreadyExists(Exception):
    pass


class MobileDeviceQuerySet(SmartQuerySet):

    def activate_or_create(self, uuid, user, user_agent='', name=None, slug=None, is_primary=True):
        """
        This method must be called when user is authenticated.
        It creates a new MobileDevice with auto generated token for the device and returns token.
        If MobileDevice with same UUID exists MobileDeviceAlreadyExists is raised.
        """
        secret_password = generate_key(length=settings.MOBILE_DEVICE_SECRET_PASSWORD_LENGTH)
        defaults = {
            'login_token': make_password(secret_password),
            'user_agent': user_agent[:256],
            'is_active': True,
            'name': name,
            'slug': slug,
            'is_primary': is_primary
        }
        mobile_device, is_created_mobile_device = self.get_or_create(
            uuid=uuid,
            user=user,
            defaults=defaults
        )
        mobile_device.secret_password = secret_password
        if not is_created_mobile_device and not mobile_device.is_active:
            return mobile_device.change_and_save(**defaults)
        elif is_created_mobile_device:
            return mobile_device
        else:
            raise MobileDeviceAlreadyExists('Device key already exists')


class MobileDevice(SmartModel):
    """Model used to authenticate mobile devices. Unhashed login_token is stored
    in the device keychain and serve as password to log in together with UUID via DeviceBackend."""

    # this is not UUIDField because of the strict length limitation
    uuid = models.CharField(
        verbose_name=_('UUID'),
        max_length=36,
        null=False,
        blank=False,
        db_index=True
    )
    name = models.CharField(
        verbose_name=_('name'),
        max_length=250,
        null=True,
        blank=True
    )
    slug = models.SlugField(
        verbose_name=_('slug'),
        null=True,
        blank=True
    )
    last_login = models.DateTimeField(
        verbose_name=_('last login'),
        null=True,
        blank=True
    )
    user = models.ForeignKey(
        to=django_settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        verbose_name=_('user'),
        related_name='mobile_devices'
    )
    login_token = models.CharField(
        max_length=128,
        verbose_name=_('login token'),
        null=False,
        blank=False
    )
    is_active = models.BooleanField(
        verbose_name=_('is active'),
        default=True
    )
    user_agent = models.CharField(
        verbose_name=_('user agent'),
        max_length=256,
        null=True,
        blank=True,
    )
    is_primary = models.BooleanField(
        verbose_name=_('is primary'),
        default=False
    )

    class Meta:
        unique_together = ('uuid', 'user')
        verbose_name = _('mobile device')
        verbose_name_plural = _('mobile devices')

    objects = MobileDeviceQuerySet.as_manager()

    @transaction.atomic()
    def _pre_save(self, changed, changed_fields, *args, **kwargs):
        super()._pre_save(self, changed, changed_fields, *args, **kwargs)
        if self.is_primary and 'is_primary' in self.changed_fields:
            # only one device can be primary
            self.__class__.objects.filter(user=self.user).exclude(pk=self.pk).select_for_update().change_and_save(
                is_primary=False
            )

    def _post_save(self, changed, changed_fields, *args, **kwargs):
        super()._post_save(changed, changed_fields, *args, **kwargs)
        if 'is_active' in changed_fields and not self.is_active:
            self.authorization_tokens.update(is_active=False)

    def set_login_token(self, raw_token):
        self.login_token = make_password(raw_token)

    def check_login_token(self, raw_token):
        def setter(raw_token):
            self.set_login_token(raw_token)
            self.save(update_fields=['login_token'])
        return check_password(raw_token, self.login_token, setter)


class OneTimePasswordManager(BaseHashKeyManager):

    def _hash_key(self, key, slug, **kwargs):
        return hash_key(key, salt=slug)

    def create(self, slug, key_generator, **kwargs):
        return super().create(key_generator, slug=slug, **kwargs)


class OneTimePassword(SmartModel):
    """
    Specific verification tokens that can be send via e-mail, SMS or another transmission medium
    to check user authorization (example password reset)
    """

    key = models.CharField(
        verbose_name=_('key'),
        max_length=128,
        unique=True,
        db_index=True,
        null=False,
        blank=False
    )
    expires_at = models.DateTimeField(
        verbose_name=_('expires at'),
        null=True,
        blank=True,
    )
    slug = models.SlugField(
        verbose_name=_('slug'),
        null=False,
        blank=False
    )
    is_active = models.BooleanField(
        verbose_name=_('is active'),
        default=True
    )
    data = models.JSONField(
        verbose_name=_('data'),
        null=True,
        blank=True,
        encoder=DjangoJSONEncoder
    )
    related_objects = GenericManyToManyField()

    secret_key = None

    objects = OneTimePasswordManager()

    class Meta:
        ordering = ('-created_at',)
        verbose_name = _('one time password')
        verbose_name_plural = _('one time passwords')

    @property
    def is_expired(self):
        return self.expires_at and self.expires_at < timezone.now()


class AuthorizationRequest(SmartModel):

    authorization_token = models.ForeignKey(
        verbose_name=_('authorization token'),
        to=AuthorizationToken,
        related_name='authorization_requests',
        null=True,
        blank=True,
        on_delete=models.SET_NULL
    )
    user = models.ForeignKey(
        verbose_name=_('user'),
        to=django_settings.AUTH_USER_MODEL,
        related_name='authorization_requests',
        null=False,
        blank=False,
        on_delete=models.CASCADE
    )
    slug = models.SlugField(
        verbose_name=_('slug'),
        null=True,
        blank=True
    )
    backend = models.CharField(
        verbose_name=_('backend'),
        max_length=250,
        null=False,
        blank=False
    )
    title = models.CharField(
        verbose_name=_('title'),
        max_length=250,
        null=False,
        blank=False
    )
    description = models.TextField(
        verbose_name=_('description'),
        null=True,
        blank=True
    )
    result = IntegerEnumField(
        verbose_name=_('result'),
        enum=AuthorizationRequestResult,
        null=True,
        blank=True
    )
    data = models.JSONField(
        verbose_name=_('data'),
        null=True,
        blank=True,
        encoder=DjangoJSONEncoder
    )
    expires_at = models.DateTimeField(
        verbose_name=_('expires at'),
        null=True,
        blank=True,
    )
    granted_at = models.DateTimeField(
        verbose_name=_('granted at'),
        null=True,
        blank=True
    )
    related_objects = GenericManyToManyField()

    class Meta:
        ordering = ('-created_at',)
        verbose_name = _('authorization request')
        verbose_name_plural = _('authorization requests')

    @property
    def is_expired(self):
        return self.expires_at < timezone.now()

    @property
    def state(self):
        if not self.result and self.is_expired:
            return AuthorizationRequestState.EXPIRED
        elif not self.result:
            return AuthorizationRequestState.WAITING
        else:
            return AuthorizationRequestState[self.result.name]



from auth_token.contrib.common.auth_security.validators import LOGIN_THROTTLING_VALIDATORS
from auth_token.contrib.is_core_auth.default.resource import AuthResource as DefaultAuthResource

from security.enums import InputRequestSlug
from security.decorators import throttling
from security.utils import update_logged_request_data


class AuthResource(DefaultAuthResource):

    @throttling(*LOGIN_THROTTLING_VALIDATORS)
    def post(self):
        return super().post()

    def _sucessful_login(self, request):
        update_logged_request_data(self.request, slug=InputRequestSlug.SUCCESSFUL_LOGIN_REQUEST)
        return super()._sucessful_login(request)

    def _unsucessful_login(self, request):
        update_logged_request_data(self.request, slug=InputRequestSlug.UNSUCCESSFUL_LOGIN_REQUEST)
        return super()._unsucessful_login(request)



from django.contrib.auth.views import LoginView

from auth_token.contrib.common.default.views import TokenLoginView as DefaultTokenLoginView

from security.decorators import throttling_all
from security.enums import InputRequestSlug
from security.utils import update_logged_request_data

from .validators import LOGIN_THROTTLING_VALIDATORS


@throttling_all(*LOGIN_THROTTLING_VALIDATORS)
class TokenLoginView(DefaultTokenLoginView):

    def form_valid(self, form):
        update_logged_request_data(self.request, slug=InputRequestSlug.SUCCESSFUL_LOGIN_REQUEST)
        return super().form_valid(form)

    def form_invalid(self, form):
        update_logged_request_data(self.request, slug=InputRequestSlug.UNSUCCESSFUL_LOGIN_REQUEST)
        return super().form_invalid(form)


@throttling_all(*LOGIN_THROTTLING_VALIDATORS)
class LoginCodeVerificationView(LoginView):
    pass


class InputLogMixin:

    def log_successful_request(self):
        update_logged_request_data(self.request, slug=InputRequestSlug.SUCCESSFUL_2FA_CODE_VERIFICATION_REQUEST)

    def log_unsuccessful_request(self):
        update_logged_request_data(self.request, slug=InputRequestSlug.UNSUCCESSFUL_2FA_CODE_VERIFICATION_REQUEST)



from django.contrib import messages
from django.contrib.auth import authenticate
from django.contrib.auth.views import redirect_to_login
from django.core.exceptions import ImproperlyConfigured
from django.http import Http404, HttpResponseRedirect
from django.utils.decorators import method_decorator
from django.utils.translation import gettext
from django.views.decorators.csrf import csrf_exempt
from django.views.generic.base import RedirectView, View

from auth_token.config import settings
from auth_token.utils import login

from .helpers import Protocol, acquire_token_by_auth_code_flow, get_sign_in_flow, init_saml_auth


def _check_session(request):
    if not hasattr(request, 'session'):
        raise ImproperlyConfigured('Django SessionMiddleware must be enabled to use MS SSO')


class ProtocolCheckMixin:

    def dispatch(self, request, *args, **kwargs):
        if self.protocol != Protocol(settings.MS_SSO_PROTOCOL):
            raise Http404
        return super().dispatch(request, *args, **kwargs)


class OauthMsLogin(ProtocolCheckMixin, RedirectView):

    protocol = Protocol.OAUTH

    def get_redirect_url(self, *args, **kwargs):
        sign_flow = get_sign_in_flow()
        sign_flow['next'] = self.request.GET.get('next', '/')

        _check_session(self.request)
        self.request.session['auth_token_ms_sso_auth_flow'] = sign_flow
        return sign_flow['auth_uri']


class SamlMsLogin(ProtocolCheckMixin, RedirectView):

    protocol = Protocol.SAML

    def get_redirect_url(self, *args, **kwargs):
        return init_saml_auth(self.request).login(return_to=self.request.GET.get('next', '/'))


class BaseMsCallback(ProtocolCheckMixin, View):

    allowed_cookie = True
    allowed_header = False

    def _get_next(self):
        return '/'

    def _redirect_to_login_with_error(self):
        messages.error(
            self.request, gettext('Microsoft SSO login was unsuccessful, please use another login method')
        )
        return redirect_to_login(self._get_next())

    def _do_login(self, **kwargs):
        user = authenticate(self.request, **kwargs)
        if not user:
            return self._redirect_to_login_with_error()
        else:
            login(
                self.request,
                user,
                allowed_cookie=self.allowed_cookie,
                allowed_header=self.allowed_header,
                two_factor_login=False
            )
            return HttpResponseRedirect(self._get_next())


class OauthMsCallback(BaseMsCallback):

    protocol = Protocol.OAUTH

    def _get_next(self):
        return self.sign_flow['next'] if self.sign_flow else super()._get_next()

    def get(self, *args, **kwargs):
        _check_session(self.request)
        self.sign_flow = self.request.session.get('auth_token_ms_sso_auth_flow')
        if not self.sign_flow:
            return self._redirect_to_login_with_error()

        result = acquire_token_by_auth_code_flow(self.sign_flow, self.request.GET)
        if 'access_token' not in result:
            return self._redirect_to_login_with_error()

        return self._do_login(mso_token=result['access_token'])


@method_decorator(csrf_exempt, name='dispatch')  # POST request from external service, we must disable CSRF
class SamlMsCallback(BaseMsCallback):

    protocol = Protocol.SAML

    def _get_next(self):
        relay_state = self.request.POST.get('RelayState')
        # Test that value starts with "/" (relative redirect) to avoid open redirect attack.
        # See: https://cwe.mitre.org/data/definitions/601.html
        return relay_state if relay_state and relay_state.startswith('/') else super()._get_next()

    def post(self, *args, **kwargs):
        return self._do_login(using_saml=True)



from auth_token.utils import is_installed_security

if is_installed_security:
    from .auth_security.resource import AuthResource  # noqa: F401
else:
    from .default.resource import AuthResource  # noqa: F401



from django.urls import path

from .views import OauthMsCallback, OauthMsLogin, SamlMsCallback, SamlMsLogin


urlpatterns = [
    path(
        'login/mso',
        OauthMsLogin.as_view(),
        name='ms-sso-login',
    ),
    path(
        'login/mso/callback',
        OauthMsCallback.as_view(),
        name='ms-sso-redirect',
    ),
    path(
        'login/mso/saml',
        SamlMsLogin.as_view(),
        name='ms-sso-saml-login',
    ),
    path(
        'login/mso/saml/callback',
        SamlMsCallback.as_view(),
        name='ms-sso-saml-redirect',
    ),
]



import time

from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend
from django.core.cache import cache

from auth_token.contrib.ms_sso.views import init_saml_auth

from .helpers import get_user_data


class BaseMsSsoBackend(ModelBackend):
    """
    Base class for MS SSO related backends.
    """

    user_model = get_user_model()

    def _get_natural_key(self):
        """
        This method should return user's identifier (e.g. e-mail) received from the identity provider.
        """
        raise NotImplementedError

    def _get_user_from_natural_key(self):
        try:
            return self.user_model._default_manager.get_by_natural_key(self._get_natural_key())
        except self.user_model.DoesNotExist:
            return None

    def authenticate(self, request, **kwargs):
        if (user := self._get_user_from_natural_key()) and self.user_can_authenticate(user):
            return user
        else:
            return None


class OauthMsSsoBackend(BaseMsSsoBackend):

    def _get_natural_key(self):
        return self.ms_user_data['userPrincipalName']

    def authenticate(self, request, mso_token, **kwargs):
        self.ms_user_data = get_user_data(mso_token)

        if not self.ms_user_data:
            return None

        return super().authenticate(request)


class SamlMsSsoBackend(BaseMsSsoBackend):

    def _get_natural_key(self):
        auth = init_saml_auth(self.request)
        auth.process_response()

        # To avoid replay attacks, check if the message was already processed
        cache_key = f'django_auth_token_saml_message_{auth.get_last_message_id()}_{auth.get_last_assertion_id()}'
        if cache.get(cache_key):
            return None
        cache.set(cache_key, True, timeout=auth.get_last_assertion_not_on_or_after() - int(time.time()))

        return auth.get_nameid()

    def authenticate(self, request, using_saml, **kwargs):
        self.request = request
        return super().authenticate(request)



import time
from datetime import timedelta
from unittest.mock import MagicMock, patch

from django.contrib.auth.models import User
from django.core.cache import cache
from django.test import override_settings
from django.test.client import RequestFactory
from django.utils.timezone import localtime, now

import httpretty
import responses
from freezegun import freeze_time
from germanium.test_cases.client import ClientTestCase
from germanium.tools import assert_equal, assert_false, assert_is_none, assert_not_equal, assert_true

from auth_token.config import settings
from auth_token.contrib.ms_sso.backends import OauthMsSsoBackend
from auth_token.contrib.ms_sso.helpers import get_ms_sso_login_url, get_user_data, init_saml_auth

from .base import BaseTestCaseMixin


__all__ = (
    'OauthMsSsoTestCase',
    'SamlMsSsoTestCase',
)


class OauthMsSsoTestCase(BaseTestCaseMixin, ClientTestCase):

    def test_get_user_data_should_return_none_for_invalid_ms_response(self):
        with responses.RequestsMock() as resp:
            resp.add(
                responses.GET, 'https://graph.microsoft.com/v1.0/me', status=401,
                content_type='application/json', json={
                    'error': {
                        'code': 'InvalidAuthenticationToken',
                        'message': 'Access token is empty.',
                        'innerError': {
                            'date': '2022-12-28T12:39:11',
                            'request-id': '373332eb-b369-4577-8c1b-62c075a22926',
                            'client-request-id': '373332eb-b369-4577-8c1b-62c075a22926'
                        }
                    }
                }
            )
            assert_is_none(get_user_data('token'))

    def test_get_user_data_should_return_user_data(self):
        with responses.RequestsMock() as resp:
            user_data = {
               'displayName': 'Adele Vance',
               'mail': 'AdeleV@contoso.onmicrosoft.com',
               'userPrincipalName': 'AdeleV@contoso.onmicrosoft.com',
               'id': '87d349ed-44d7-43e1-9a83-5f2406dee5bd'
            }
            resp.add(
                responses.GET, 'https://graph.microsoft.com/v1.0/me', status=200,
                content_type='application/json', json=user_data
            )
            assert_equal(get_user_data('token'), user_data)

    def test_ms_sso_backend_should_not_authenticate_not_logged_user(self):
        with responses.RequestsMock() as resp:
            resp.add(responses.GET, 'https://graph.microsoft.com/v1.0/me', status=401, content_type='application/json')
            assert_is_none(OauthMsSsoBackend().authenticate(None, 'token'))

    def test_ms_sso_backend_should_return_none_for_none_token(self):
        assert_is_none(OauthMsSsoBackend().authenticate(None, None))

    def test_ms_sso_backend_should_return_none_for_not_existing_user(self):
        with responses.RequestsMock() as resp:
            user_data = {
                'displayName': 'Test Test',
                'mail': 'test@localhost',
                'userPrincipalName': 'test',
                'id': '87d349ed-44d7-43e1-9a83-5f2406dee5bd'
            }
            resp.add(
                responses.GET, 'https://graph.microsoft.com/v1.0/me', status=200,
                content_type='application/json', json=user_data
            )
            assert_is_none(OauthMsSsoBackend().authenticate(None, 'token'))

    def test_ms_sso_backend_should_return_the_right_user(self):
        user = self.create_user()
        with responses.RequestsMock() as resp:
            user_data = {
                'displayName': 'Test Test',
                'mail': 'test@localhost',
                'userPrincipalName': 'test',
                'id': '87d349ed-44d7-43e1-9a83-5f2406dee5bd'
            }
            resp.add(
                responses.GET, 'https://graph.microsoft.com/v1.0/me', status=200,
                content_type='application/json', json=user_data
            )
            assert_equal(OauthMsSsoBackend().authenticate(None, 'token'), user)

    def test_login_mso_callback_should_log_user(self):
        user = self.create_user()
        with patch('auth_token.contrib.ms_sso.views.get_sign_in_flow') as mocked_get_sign_in_flow:
            mocked_get_sign_in_flow.return_value = {
                'state': 'state',
                'redirect_uri': None,
                'scope': ['openid', 'profile', 'user.read', 'offline_access'],
                'auth_uri': 'https://login.microsoftonline.com/test/oauth2/v2.0/authorize',
                'code_verifier': 'testverifier',
                'nonce': 'testnonce',
                'claims_challenge': None
            }
            response = self.get('/login/mso')
            assert_equal(response.status_code, 302)
            assert_equal(response['location'], 'https://login.microsoftonline.com/test/oauth2/v2.0/authorize')
            with patch('auth_token.contrib.ms_sso.views.acquire_token_by_auth_code_flow') \
                    as mocked_acquire_token_by_auth_code_flow:
                mocked_acquire_token_by_auth_code_flow.return_value = {
                    'access_token': 'token'
                }
                with responses.RequestsMock() as resp:
                    user_data = {
                        'displayName': 'Test Test',
                        'mail': 'test@localhost',
                        'userPrincipalName': 'test',
                        'id': '87d349ed-44d7-43e1-9a83-5f2406dee5bd'
                    }
                    resp.add(
                        responses.GET, 'https://graph.microsoft.com/v1.0/me', status=200,
                        content_type='application/json', json=user_data
                    )
                    response = self.get('/login/mso/callback')
                    assert_equal(response.wsgi_request.user, user)
                    assert_equal(response.status_code, 302)
                    assert_equal(response['location'], '/')

    def test_login_mso_callback_without_access_token_should_not_log_user(self):
        with patch('auth_token.contrib.ms_sso.views.get_sign_in_flow') as mocked_get_sign_in_flow:
            mocked_get_sign_in_flow.return_value = {
                'state': 'state',
                'redirect_uri': None,
                'scope': ['openid', 'profile', 'user.read', 'offline_access'],
                'auth_uri': 'https://login.microsoftonline.com/test/oauth2/v2.0/authorize',
                'code_verifier': 'testverifier',
                'nonce': 'testnonce',
                'claims_challenge': None
            }
            response = self.get('/login/mso')
            assert_equal(response.status_code, 302)
            assert_equal(response['location'], 'https://login.microsoftonline.com/test/oauth2/v2.0/authorize')
            with patch('auth_token.contrib.ms_sso.views.acquire_token_by_auth_code_flow') \
                    as mocked_acquire_token_by_auth_code_flow:
                mocked_acquire_token_by_auth_code_flow.return_value = {}
                response = self.get('/login/mso/callback')
                assert_false(response.wsgi_request.user.is_authenticated)
                assert_equal(response.status_code, 302)
                assert_equal(response['location'], '/accounts/login/?next=/')

    def test_login_mso_callback_without_sign_flow_should_not_log_user(self):
        response = self.get('/login/mso/callback')
        assert_false(response.wsgi_request.user.is_authenticated)
        assert_equal(response.status_code, 302)
        assert_equal(response['location'], '/accounts/login/?next=/')

    def test_get_ms_sso_login_url_should_return_correct_url_for_oauth(self):
        assert_equal(get_ms_sso_login_url(), '/login/mso')

    @override_settings(AUTH_TOKEN_MS_SSO_PROTOCOL='saml')
    def test_endpoints_for_oauth_should_return_404_if_this_protocol_is_not_set(self):
        assert_equal(self.c.get('/login/mso').status_code, 404)
        assert_equal(self.c.get('/login/mso/callback').status_code, 404)


@override_settings(AUTH_TOKEN_MS_SSO_PROTOCOL='saml')
@override_settings(AUTH_TOKEN_MS_SSO_SAML_METADATA_URL='http://localhost/metadata')
@override_settings(AUTH_TOKEN_MS_SSO_SAML_ENTITY_ID='My Test App')
class SamlMsSsoTestCase(BaseTestCaseMixin, ClientTestCase):

    def _get_auth_mock(self):
        auth_mock = MagicMock()
        auth_mock.get_nameid.return_value = 'john@doe.com'
        auth_mock.get_last_message_id.return_value = '123'
        auth_mock.get_last_assertion_id.return_value = '456'
        auth_mock.get_last_assertion_not_on_or_after.return_value = int(time.time()) + 3600  # valid for 1 hour
        return auth_mock

    def _register_metadata_url(self, httpretty):
        with open('dj/apps/app/tests/saml_metadata_response.xml', 'r') as f:
            saml_metadata_response = f.read()

        httpretty.register_uri(
            httpretty.GET,
            'http://localhost/metadata',
            status=200,
            content_type='application/xml',
            body=saml_metadata_response,
        )

    @httpretty.activate  # we cannot use responses, becuase python3-saml uses urllib, not requests
    def test_saml_should_log_in_user(self):
        CACHE_KEY = 'django_auth_token_saml_message_123_456'
        user = self.create_user(username='john@doe.com')

        self._register_metadata_url(httpretty)

        # verify we're redirected to the identity provider
        response = self.get('/login/mso/saml?next=/dashboard')
        assert_equal(response.status_code, 302)
        assert_true(response['location'].startswith(
            'https://login.microsoftonline.com/54311acb-4c14-4e2a-ba36-fcd4de0bffa2/saml2'
        ))

        # POST payload is dynamic and signed, not easy to mock, therefore we mock auth object instead
        with patch('auth_token.contrib.ms_sso.backends.init_saml_auth') as init_mock:
            init_mock.return_value = self._get_auth_mock()
            response = self.post('/login/mso/saml/callback', data={'RelayState': ['/dashboard']})
            assert_equal(response.status_code, 302)
            assert_equal(response['location'], '/dashboard')
            assert_equal(response.wsgi_request.user, user)
            assert_true(response.wsgi_request.user.is_authenticated)
            assert_true(cache.get(CACHE_KEY))

            # logout user
            user.authorization_tokens.all().delete()

            # repeated request (replay attack) 1 second before timeout, must not succeed
            with freeze_time(localtime() + timedelta(minutes=59, seconds=59), tick=True):
                response = self.post('/login/mso/saml/callback', data={'RelayState': ['/dashboard']})
                assert_equal(response.status_code, 302)
                assert_equal(response['location'], '/accounts/login/?next=/dashboard')
                assert_not_equal(response.wsgi_request.user, user)
                assert_false(response.wsgi_request.user.is_authenticated)
                assert_true(cache.get(CACHE_KEY))

            # after timeout, cache entry is evicted
            with freeze_time(localtime() + timedelta(minutes=60), tick=True):
                assert_false(cache.get(CACHE_KEY))

        # cleanup
        cache.delete(CACHE_KEY)

    def test_saml_should_not_log_in_user(self):
        self.create_user(username='alice@doe.com')
        self.create_user(username='john@doe.com', is_active=False)
        for user in User.objects.all():
            with patch('auth_token.contrib.ms_sso.backends.SamlMsSsoBackend._get_natural_key') as mock:
                mock.return_value = 'john@doe.com'
                response = self.post('/login/mso/saml/callback', data={'RelayState': ['/dashboard']})
                assert_equal(response.status_code, 302)
                assert_equal(response['location'], '/accounts/login/?next=/dashboard')
                assert_not_equal(response.wsgi_request.user, user)
                assert_false(response.wsgi_request.user.is_authenticated)

    def test_get_ms_sso_login_url_should_return_correct_url_for_saml(self):
        assert_equal(get_ms_sso_login_url(), '/login/mso/saml')

    @override_settings(AUTH_TOKEN_MS_SSO_PROTOCOL='oauth')
    def test_endpoints_for_saml_should_return_404_if_this_protocol_is_not_set(self):
        assert_equal(self.c.get('/login/mso/saml').status_code, 404)
        assert_equal(self.c.get('/login/mso/saml/callback').status_code, 404)

    def test_saml_should_prevent_open_redirect_attack(self):
        user = self.create_user(username='john@doe.com')
        with patch('auth_token.contrib.ms_sso.backends.SamlMsSsoBackend._get_natural_key') as mock:
            mock.return_value = 'john@doe.com'
            response = self.post('/login/mso/saml/callback', data={'RelayState': ['http://example.com/dashboard']})
            assert_equal(response.status_code, 302)
            assert_equal(response['location'], '/')  # absolute URL is ignored, fallback to '/'
            assert_equal(response.wsgi_request.user, user)
            assert_true(response.wsgi_request.user.is_authenticated)

    @httpretty.activate
    def test_saml_auth_should_be_correctly_initialized(self):
        self._register_metadata_url(httpretty)
        auth = init_saml_auth(RequestFactory().get('/'))
        service_provider_settings = auth._settings._sp
        assert_equal(service_provider_settings['entityId'], 'My Test App')
        assert_equal(
            service_provider_settings['assertionConsumerService']['url'],
            'http://testserver/login/mso/saml/callback',
        )
        security_settings = auth._settings._security
        assert_equal(security_settings['requestedAuthnContext'], False)



# Generated by Django 3.1 on 2022-10-25 12:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('auth_token', '0015_auto_20210308_0919'),
    ]

    operations = [
        migrations.AlterField(
            model_name='mobiledevice',
            name='uuid',
            field=models.CharField(db_index=True, max_length=36, verbose_name='UUID'),
        ),
    ]



from datetime import timedelta

from django.test import override_settings
from django.utils.timezone import localtime

from freezegun import freeze_time
from germanium.decorators import data_consumer
from germanium.test_cases.rest import RestTestCase
from germanium.tools import assert_equal, assert_not_equal
from germanium.tools.http import assert_http_ok

from auth_token.models import AuthorizationToken

from .base import BaseTestCaseMixin


__all__ = (
    'MiddlewareTestCase',
)


class MiddlewareTestCase(BaseTestCaseMixin, RestTestCase):

    LOGIN_URL = '/is_core/api/login/'
    INDEX_URL = '/is_core/'

    def _get_auth_header(self, token):
        return {'HTTP_AUTHORIZATION': 'Bearer {}'.format(token)}

    @override_settings(AUTH_TOKEN_EXPIRATION_DELTA=60)
    @data_consumer('create_user')
    def test_token_middleware_should_respect_expiration_delta_setting(self, user):
        response = self.post(self.LOGIN_URL, {'username': 'test', 'password': 'test'})
        assert_equal(AuthorizationToken.objects.count(), 1)

        token = response.json()['token']
        token_obj = AuthorizationToken.objects.last()
        expires_at = token_obj.expires_at

        with freeze_time(localtime() + timedelta(seconds=59)):
            self.get(self.INDEX_URL, headers=self._get_auth_header(token))
            assert_equal(expires_at, token_obj.refresh_from_db().expires_at)

        with freeze_time(localtime() + timedelta(seconds=60)):
            self.get(self.INDEX_URL, headers=self._get_auth_header(token))
            assert_not_equal(expires_at, token_obj.refresh_from_db().expires_at)
